-- Iriel's Ring Template
----------------------------------------------------------------------
--
-- Template code for partial ring display
--
-- Divides the ring up into four quadrants:
--   Q1-TopRight Q2-BottomRight Q3-BottomLeft Q4-TopLeft
--
-- Other terminology:
--   self.radius - The outer radius of the ring (Also the texture width)
--   self.ringFactor - The ratio of inner radius/outer radius

---------------------------------------------------------------------------
-- QUADRANT MAPPING FUNCTIONS
--   The subsetting code is all written relative to the first quandrant, but
--   each quadrant has a pair of functions to manage mapping location and
--   texture operations from that 'normalized' coordinate system into the
--   appropriate one for the quadrant in question
--
-- SUBSET FUNCTION
--   This function displays a subset of a quadrant, using a subset of a
--   texture (The texture subsetting is optional, so the same function
--   can be used for slice stretching). Note this does not issue a
--   ClearAllPoints.
--
--   setSubsetFuncs[quadrant](tex, parname, radius, xlo, xhi, ylo, yhi, TM)
--
--   Parameters:
--     tex      - The texture object
--     parname  - The name of the texture's parent (for SetPoint use)
--     radius   - The outer radius of the ring (i.e. texture width & height)
--     xlo, xhi - X Coordinate bounds (from center of the ring)
--     ylo, yhi - Y Coordinate bounds (From center of the ring)
--     TM       - (OPTIONAL) If present then use this as a mapping function
--                for texCoords, and set them (otherwise leave untouched)
--

local RING_TEMPLATE_COMPATIBILITY = 2; -- Increment on API changes
local RING_TEMPLATE_SUB_VERSION = 7;   -- Increment on all changes

local function GetTemplateInstance() 
   local INSTANCE = {};
   local METHODS = {};
   INSTANCE.methods = METHODS;
   
   local initSetSubsetFuncs = {};
   local initTexCoordMapFuncs = {};

   -- Texture coordinate mapping functions
   initTexCoordMapFuncs[1] = function(ulx,uly, llx,lly, urx,ury, lrx,lry) 
			    return ulx,uly, llx,lly, urx,ury, lrx,lry;
			 end
   initTexCoordMapFuncs[2] = function(ulx,uly, llx,lly, urx,ury, lrx,lry) 
			    return llx,lly, lrx,lry, ulx, uly, urx,ury;
			 end
   initTexCoordMapFuncs[3] = function(ulx,uly, llx,lly, urx,ury, lrx,lry) 
			    return lrx,lry, urx,ury, llx,lly, ulx,uly;
			 end
   initTexCoordMapFuncs[4] = function(ulx,uly, llx,lly, urx,ury, lrx,lry) 
			    return urx,ury, ulx,uly, lrx,lry, llx,lly;
			 end

   METHODS.ringTexCoordMapFuncs = initTexCoordMapFuncs;
   METHODS.sliceTexCoordMapFuncs = initTexCoordMapFuncs;

   -- Q1: TOP RIGHT
   initSetSubsetFuncs[1] = 
      function(tex, parname, radius, xlo, xhi, ylo, yhi, TM)
	 if (TM) then
	    tex:SetTexCoord(TM(xlo,1-yhi, xlo,1-ylo, xhi,1-yhi, xhi,1-ylo));
	 end
	 tex:SetPoint("TOPLEFT", parname, "CENTER", xlo*radius, yhi*radius);
	 tex:SetPoint("BOTTOMRIGHT", parname, "CENTER", xhi*radius, ylo*radius);
      end

   -- Q2: BOTTOM RIGHT
   initSetSubsetFuncs[2] =
      function(tex, parname, radius, xlo, xhi, ylo, yhi, TM)
	 if (TM) then
	    tex:SetTexCoord(TM(xlo,1-yhi, xlo,1-ylo, xhi,1-yhi, xhi,1-ylo));
	 end
	 tex:SetPoint("TOPRIGHT", parname, "CENTER", yhi*radius, -xlo*radius);
	 tex:SetPoint("BOTTOMLEFT", parname, "CENTER", ylo*radius, -xhi*radius);
      end

   -- Q3: BOTTOM LEFT
   initSetSubsetFuncs[3] = 
      function(tex, parname, radius, xlo, xhi, ylo, yhi, TM)
	 if (TM) then
	    tex:SetTexCoord(TM(xlo,1-yhi, xlo,1-ylo, xhi,1-yhi, xhi,1-ylo));
	 end
	 tex:SetPoint("BOTTOMRIGHT", parname, "CENTER", -xlo*radius, -yhi*radius);
	 tex:SetPoint("TOPLEFT", parname, "CENTER", -xhi*radius, -ylo*radius);
      end

   -- Q4: TOP LEFT
   initSetSubsetFuncs[4] = 
      function(tex, parname, radius, xlo, xhi, ylo, yhi, TM)
	 if (TM) then
	    tex:SetTexCoord(TM(xlo,1-yhi, xlo,1-ylo, xhi,1-yhi, xhi,1-ylo));
	 end
	 tex:SetPoint("BOTTOMLEFT", parname, "CENTER", -yhi*radius, xlo*radius);
	 tex:SetPoint("TOPRIGHT", parname, "CENTER", -ylo*radius, xhi*radius);
      end

   METHODS.setSubsetFuncs = initSetSubsetFuncs;

   -- The 'Work' function, which handles subset rendering for a single
   -- quadrant (normalized to Q1) that starts at the 0 degree position
   --
   -- Params:
   --  self - The ring template instance
   --  A    - The angle within the quadrant (degrees, 0 <= A < 90)
   --  T    - The main texture for the quadrant
   --  SS   - The texture subset mapping function for the quadrant
   --  TM   - The ring texture coord mapping function for the quadrant
   --  S    - Slice texture (optional)
   --  C    - Chip texture (optional)
   function METHODS.DoQuadrant(self, A, T, SS, TM, S, C)
      -- Grab local references to important textures
      S = S or self.slice;
      C = C or self.chip;

      -- If no part of this quadrant is visible, just hide all the textures
      -- and be done.
      if (A == 0) then
	 T:Hide();
	 C:Hide();
	 S:Hide();
	 return;
      end

      -- More local references, grab the ring dimensions, and the frame name.
      local RF = self.ringFactor;
      local OR = self.radius;
      local name = self.name;

      -- Drawing scheme uses three locations
      --   N (Nx,Ny) - The 'Noon' position (Nx=0, Ny=1)
      --   O (Ox,Oy) - Intersection of angle line with Outer edge
      --   I (Ix,Iy) - Intersection of angle line with Inner edge

      -- Calculated locations:
      --   Arad  - Angle in radians
      --   Ox,Oy - O coordinates
      --   Ix,Iy - I coordinates
      local Arad = math.rad(A);
      local Ox = math.sin(Arad);
      local Oy = math.cos(Arad);
      local Ix = Ox * RF;
      local Iy = Oy * RF;

      -- Treat first and last halves differently to maximize size of main
      -- texture subset.
      if (A <= 45) then
	 -- Main subset is from N to I
	 SS(T, name, OR, 0, Ix, Iy, 1,  TM);
	 -- Chip is subset from (Ix,Oy) to (Ox,Ny) (Right edge of main)
	 SS(C, name, OR, Ix, Ox, Oy, 1, TM);
      else
	 -- Main subset is from N to O
	 SS(T, name, OR, 0, Ox, Oy, 1,  TM);
	 -- Chip is subset from (Nx,Iy) to (Ix,Oy) (Bottom edge of main)
	 SS(C, name, OR, 0, Ix, Iy, Oy, TM);
      end
      -- Strech slice between I and O
      SS(S, name, OR, Ix, Ox, Iy, Oy);
      -- All three textures are visible
      T:Show();
      C:Show();
      S:Show();
   end

   -- The 'Work' function, which handles subset rendering for a single
   -- quadrant (normalized to Q1) in reverse
   --
   -- Params:
   --  self - The ring template instance
   --  A    - The angle within the quadrant (degrees, 0 <= A < 90)
   --  T    - The main texture for the quadrant
   --  SS   - The texture subset mapping function for the quadrant
   --  TM   - The ring texture coord mapping function for the quadrant
   --  S    - Slice texture <Reversed> (optional)
   --  C    - Chip texture (optional)
   function METHODS.DoQuadrantReversed(self, A, T, SS, TM, S, C)
      -- Grab local references to important textures
      S = S or self.slice;
      C = C or self.chip;

      -- If no part of this quadrant is visible, just hide all the textures
      -- and be done.
      if (A == 0) then
	 T:Hide();
	 C:Hide();
	 S:Hide();
	 return;
      end

      -- More local references, grab the ring dimensions, and the frame name.
      local RF = self.ringFactor;
      local OR = self.radius;
      local name = self.name;

      -- Drawing scheme uses three locations
      --   E (Ex,Ey) - The 'End' position (Nx=1, Ny=0)
      --   O (Ox,Oy) - Intersection of angle line with Outer edge
      --   I (Ix,Iy) - Intersection of angle line with Inner edge

      -- Calculated locations:
      --   Arad  - Angle in radians
      --   Ox,Oy - O coordinates
      --   Ix,Iy - I coordinates
      local Arad = math.rad(A);
      local Ox = math.cos(Arad);
      local Oy = math.sin(Arad);
      local Ix = Ox * RF;
      local Iy = Oy * RF;

      -- Treat first and last halves differently to maximize size of main
      -- texture subset.
      if (A <= 45) then
	 -- Main subset is from N to I
	 SS(T, name, OR, Ix, 1,  0, Iy, TM);
	 -- Chip is subset from (Ix,Oy) to (Ox,Ny) (Right edge of main)
	 SS(C, name, OR, Ox, 1, Iy, Oy, TM);
      else
	 -- Main subset is from N to O
	 SS(T, name, OR, Ox, 1,  0, Oy, TM);
	 -- Chip is subset from (Nx,Iy) to (Ix,Oy) (Bottom edge of main)
	 SS(C, name, OR, Ix, Ox, 0, Iy, TM);
      end
      -- Strech slice between I and O
      SS(S, name, OR, Ix, Ox, Iy, Oy);
      -- All three textures are visible
      T:Show();
      C:Show();
      S:Show();
   end

   -- The 'Work' function, which handles subset rendering for a single
   -- quadrant (normalized to Q1), subsetted between two angle
   --
   -- Params:
   --  self - The ring template instance
   --  BA   - The BEGIN angle within the quadrant (degrees, 0 <= BA <= EA)
   --  EA   - The END  angle within the quadrant (degrees, BA <= EA <= 90)
   --  T    - The main texture for the quadrant
   --  SS   - The texture subset mapping function for the quadrant
   --  TM   - The ring texture coord mapping function for the quadrant
   --  BS   - Begin Slice texture (optional)
   --  BC   - Begin Chip texture (optional)
   --  ES   - End Slice texture (optional)
   --  EC   - End Chip texture (optional)
   --  LL   - Line piece (optional)
   --  LTM  - Line piece texture map (optional)
   function METHODS.DoQuadrantPartial(self, BA, EA, T, SS, TM,
						    BS, BC, ES, EC, LL, LTM)
      -- Grab local references to important textures
      BS = BS or self.slice2;
      BC = BC or self.chip2;
      ES = ES or self.slice;
      EC = EC or self.chip;
      LL = LL or self.quadExtra;
      LTM = LTM or TM;

      -- If no part of this quadrant is visible, just hide all the textures
      -- and be done.
      if (BA == EA) then
	 T:Hide();
	 BC:Hide();
	 BS:Hide();
	 EC:Hide();
	 ES:Hide();
	 LL:Hide();
	 return;
      end

      -- Simplifications for common cases
      if (BA == 0) then
	 LL:Hide();
	 BC:Hide();
	 BS:Hide();
	 return self:DoQuadrant(EA, T, SS, TM, ES, EC);
      elseif (EA == 90) then
	 LL:Hide();
	 EC:Hide();
	 ES:Hide();
	 return self:DoQuadrantReversed(90-BA, T, SS, TM, BS, BC);
      end

      -- More local references, grab the ring dimensions, and the frame name.
      local RF = self.ringFactor;
      local OR = self.radius;
      local name = self.name;

      -- Drawing scheme uses four locations
      --   BO (BOx,EOy) - Intersection of begin angle line with Outer edge
      --   BI (BIx,EIy) - Intersection of begin angle line with Inner edge
      --   EO (EOx,EOy) - Intersection of end angle line with Outer edge
      --   EI (EIx,EIy) - Intersection of end angle line with Inner edge

      -- Calculated locations:
      local BArad = math.rad(BA);
      local BOx = math.sin(BArad);
      local BOy = math.cos(BArad);
      local BIx = BOx * RF;
      local BIy = BOy * RF;
      local EArad = math.rad(EA);
      local EOx = math.sin(EArad);
      local EOy = math.cos(EArad);
      local EIx = EOx * RF;
      local EIy = EOy * RF;

      -- Test for 'small segment' special case
      if (EA >= 45) then
	 -- Calculate where X=BOx crosses the E line
	 -- E line is EOx * y = EOy * x  ==>  y = (EOy * x) / EOx
	 local BOEy = (EOy * BOx) / EOx;
	 if (BOEy > BIy) then
	    -- Small segment!
	    local EIBy = (BOy * EIx) / BOx;

	    local l = (BOy - EIy) / 2;

	    LL:SetTexture(ES:GetTexture());
	    LL:SetTexCoord(LTM(1 - (l / (EIBy-EIy)),0, 
			       1,0,
			       0,1,
			       (l / (BOy-BOEy)),1)); 
	    SS(LL, name, OR, EIx, BOx, EIy, BOy);
	    
	    SS(T,  name, OR, BIx, EIx, EIy, BIy, TM);
	    SS(BC, name, OR, BOx, EOx, EOy, BOy, TM);
	    SS(BS, name, OR, BIx, EIx, BIy, EIBy);
	    SS(ES, name, OR, BOx, EOx, BOEy, EOy);
	    
	    LL:Show();
	    T:Show();
	    BC:Show();
	    BS:Show();
	    ES:Show();

	    EC:Hide();
	    return;
	 end
      else
	 -- Calculate where y=EOy crosses the B line
	 -- B line is BOx * y = BOy * x  ==>  x = (BOx * y) / BOy
	 local EOBx = (BOx * EOy) / BOy;
	 if (EOBx > EIx) then
	    -- Small segment!
	    local BIEx = (EOx * BIy) / EOy;
	    local l = (EOx-BIx) / 2;

	    LL:SetTexture(ES:GetTexture());
	    LL:SetTexCoord(LTM(1 - (l / (EOx-EOBx)),0, 
			       0,1,
			       1,0,
			       (l / (BIEx-BIx)),1)); 

	    SS(LL, name, OR, BIx, EOx, BIy, EOy);
	    SS(T,  name, OR, BIx, EIx, EIy, BIy, TM);
	    SS(BC, name, OR, BOx, EOx, EOy, BOy, TM);
	    SS(BS, name, OR, EOBx, BOx, EOy, BOy);
	    SS(ES, name, OR, EIx, BIEx, EIy, BIy);

	    LL:Show();
	    T:Show();
	    BC:Show();
	    BS:Show();
	    ES:Show();

	    EC:Hide();
	    return;
	 end
      end

      LL:Hide();
      
      -- There are a few alternatives here
      if (EA <= 45) then
	 -- All at the start, non-overlapping
	 -- Main subset is from BO to EI
	 SS(T, name, OR, BOx, EIx, EIy, BOy, TM);
	 SS(BC, name, OR, BIx, BOx, EIy, BIy, TM);
	 SS(EC, name, OR, EIx, EOx, EOy, BOy, TM);
      elseif (BA >= 45) then
	 -- All at the end, non-overlapping
	 -- Main subset is from BI to EO
	 SS(T, name, OR, BIx, EOx, EOy, BIy, TM);
	 SS(BC, name, OR, BOx, EOx, BIy, BOy, TM);
	 SS(EC, name, OR, BIx, EIx, EIy, EOy, TM);
      else
	 -- Split across, non-overlapping
	 -- Main subset is from BO to EO
	 SS(T, name, OR, BOx, EOx, EOy, BOy, TM);
	 SS(BC, name, OR, BIx, BOx, EOy, BIy, TM);
	 SS(EC, name, OR, BOx, EIx, EIy, EOy, TM);
      end

      -- Treat first and last halves differently to maximize size of main
      -- texture subset.
      if (BA <= 45) then
	 -- Main subset is from N to I
	 --   SS(T, name, OR, Ix, 1,  0, Iy);
	 -- Chip is subset from (Ix,Oy) to (Ox,Ny) (Right edge of main)
	 -- SS(C, name, OR, Ox, 1, Iy, Oy);
      else
	 -- Main subset is from N to O
	 --SS(T, name, OR, Ox, 1,  0, Oy);
	 -- Chip is subset from (Nx,Iy) to (Ix,Oy) (Bottom edge of main)
	 --SS(C, name, OR, Ix, Ox, 0, Iy);
      end
      -- Strech begin slice between BI and BO
      SS(BS, name, OR, BIx, BOx, BIy, BOy);
      -- Strech end slice between EI and EO
      SS(ES, name, OR, EIx, EOx, EIy, EOy);
      -- All three textures are visible
      T:Show();
      BC:Show();
      BS:Show();
      EC:Show();
      ES:Show();
   end

   -- Method function to set the angle to display
   --
   -- Param:
   --  self  - The ring template instance
   --  angle - The angle in degrees (0 <= angle <= 360)
   --  startAngle - The angle to start at in degrees (0 <= angle <= 360)
   --               Ignored for non-plus frames
   function METHODS.SetAngle(self, angle, startAngle)
      if (self.isPlus) then
	 return self:SetAnglePlus(angle, startAngle);
      end

      -- Bounds checking on the angle so that it's between 0 and 360 (inclusive)
      if (angle < 0) then
	 angle = 0;
      end
      if (angle > 360) then
	 angle = 360;
      end

      -- Avoid duplicate work
      if (self.angle == angle) then
	 return;
      end

      local setSubsetFuncs = self.setSubsetFuncs;
      local ringTexCoordMapFuncs = self.ringTexCoordMapFuncs;
      local sliceTexCoordMapFuncs = self.sliceTexCoordMapFuncs;


      -- Determine the quadrant, and angle within the quadrant
      -- (Quadrant 5 means 'all quadrants filled')
      local quad = math.floor(angle / 90) + 1;
      local A = math.mod(angle, 90);
      local quadOfs = self.quadOffset or 0;
      local effQuad;
      if (self.reversed) then
	 effQuad = math.mod((4-quad)+quadOfs, 4)+1;
      else
	 effQuad = math.mod(quad+quadOfs-1, 4)+1;
      end

      -- Check to see if we've changed quandrants since the last time we were
      -- called. Quadrant changes re-configure some textures.
      if (quad ~= self.lastQuad) then
	 -- Loop through all quadrants
	 local qi;
	 for i=1,4 do
	    local T=self.quadrants[i];
	    if (self.reversed) then
	       qi = math.mod((4-i)+quadOfs, 4)+1;
	    else
	       qi = math.mod(i+quadOfs-1, 4)+1;
	    end
	    if (i < quad) then
	       -- If this quadrant is full shown, then show all of the texture
	       T:ClearAllPoints();
	       setSubsetFuncs[qi](T, self.name, self.radius, 0.0, 1.0, 0.0, 1.0,
				  ringTexCoordMapFuncs[qi]);
	       
	       T:Show();
	    elseif (i == quad) then
	       -- If this quadrant is partially or fully shown, begin by
	       -- showing all of the texture. Also configure the slice
	       -- texture's orientation.
	       T:ClearAllPoints();
	       setSubsetFuncs[qi](T, self.name, self.radius, 0.0, 1.0, 0.0, 1.0,
				  ringTexCoordMapFuncs[qi]);
	       T:Show();
	       local STM;
	       if (self.reversed) then
		  STM = sliceTexCoordMapFuncs[math.mod(qi+1,4)+1];
	       else
		  STM = sliceTexCoordMapFuncs[qi];
	       end
	       self.slice:SetTexCoord(STM(0.5,0, 0.5,1, 1,0, 1,1));
	       self.chip:SetTexture(T:GetTexture());
	    else
	       -- If this quadrant is not shown at all, hide it.
	       T:Hide();
	    end
	 end

	 -- Hide the chip and slice textures, and de-anchor them (They'll be
	 -- re-anchored as necessary later).
	 self.chip:Hide();
	 self.chip:ClearAllPoints();
	 self.slice:Hide();
	 self.slice:ClearAllPoints();

	 -- Remember this for next time
	 self.lastQuad = quad;
      end

      -- Remember the angle for next time
      self.angle = angle;

      -- Extra bounds check for paranoia (also handles quad 5 case)
      if ((quad < 1) or (quad > 4)) then
	 return;
      end

      -- Get quadrant-specific elements
      local T = self.quadrants[quad];
      local SS = setSubsetFuncs[effQuad];
      local TM = ringTexCoordMapFuncs[effQuad];

      -- Call the quadrant function to do the work
      if (self.reversed) then
	 self:DoQuadrantReversed(A, T, SS, TM)
      else
	 self:DoQuadrant(A, T, SS, TM)
      end
   end

   -- Method function to set the angle to display - Plus version
   --
   -- Param:
   --  self  - The ring template instance
   --  angle - The angle in degrees (0 <= angle <= 360)
   --  startAngle - The angle to start at in degrees (0 <= angle <= 360)
   function METHODS.SetAnglePlus(self, angle, startAngle)
      if (startAngle == nil) then
	 startAngle = self.startAngle or 0;
      end
      local math = math;
      if (startAngle < 0) then
	 startAngle = 360-math.mod(-startAngle, 360);
	 if (startAngle == 360) then
	    startAngle = 0;
	 end
      elseif (startAngle > 360) then
	 startAngle = math.mod(startAngle, 360);
      end

      -- Bounds checking on the angle so that it's between 0 and 360 (inclusive)
      if (angle < 0) then
	 angle = 0;
      end
      if (angle > 360) then
	 angle = 360;
      end

      -- Avoid duplicate work
      if ((self.angle == angle) and (self.startAngle == startAngle)) then
	 return;
      end

      -- Since we have control of both ends, reversal can be dealth with here
      -- which simplifies the later code
      local startA;
      if (self.reversed) then
	 startA = math.mod(720 - (startAngle + angle), 360);
      else
	 startA = startAngle;
      end

      -- Normalize starting quadrant back into quadrant 1
      local quadOfs = (self.quadOffset or 0) + math.floor(startA / 90);
      startA = math.mod(startA, 90);

      if (quadOfs ~= self.lastQuadOffset) then
	 -- Must re-do quadrant loop if we've moved around
	 self.lastQuadOffset = quadOfs;
	 self.lastQuad = nil;
      end

      -- Calculate end quadrant
      -- (Quadrant 5 means 'wraparound')
      local quad = math.floor((angle + startA) / 90) + 1;
      local endA = math.mod((angle + startA), 90);
      local effStartQuad = quadOfs + 1;
      local effQuad = math.mod(quad+quadOfs-1, 4)+1;
      
      local setSubsetFuncs = self.setSubsetFuncs;
      local ringTexCoordMapFuncs = self.ringTexCoordMapFuncs;
      local sliceTexCoordMapFuncs = self.sliceTexCoordMapFuncs;

      -- Check to see if we've changed quandrants since the last time we were
      -- called. Quadrant changes re-configure some textures.
      if (quad ~= self.lastQuad) then
	 -- Loop through all quadrants
	 local quadCheck = math.mod(quad-1,4)+1;
	 local qi, TCM, STCM;
	 for i=1,4 do
	    qi = math.mod(i+quadOfs-1, 4)+1;
	    TCM = ringTexCoordMapFuncs[qi];
	    STCM = sliceTexCoordMapFuncs[qi];
	    local T=self.quadrants[i];
	    if (i == 1) then
	       -- This is the start quadrant, configure the 2nd slice and chip
	       self.slice2:SetTexCoord(STCM(0,0, 0,1, .5,0, .5,1));
	       self.chip2:SetTexture(T:GetTexture());
	    end

	    if (i == quadCheck) then
	       -- This is the 'end' quadrant, configure slice and chip
	       T:ClearAllPoints();
	       setSubsetFuncs[qi](T, self.name, self.radius, 0,1,0,1, TCM);
	       T:SetTexCoord(TCM(0,0,0,1,1,0,1,1));
	       T:Show();
	       self.slice:SetTexCoord(STCM(.5,0, .5,1, 1,0, 1,1))
	       self.chip:SetTexture(T:GetTexture());
	    elseif (i < quad) then
	       -- If this quadrant is full shown, then show all of the texture
	       -- It may be reduced later by start processing
	       T:ClearAllPoints();
	       setSubsetFuncs[qi](T, self.name, self.radius, 0,1,0,1, TCM);
	       T:SetTexCoord(TCM(0,0,0,1,1,0,1,1));

	       T:Show();
	    else
	       -- If this quadrant is not shown at all, hide it.
	       T:Hide();
	    end
	 end

	 -- Hide the chip and slice textures, and de-anchor them (They'll be
	 -- re-anchored as necessary later).
	 self.chip:Hide();
	 self.chip:ClearAllPoints();
	 self.slice:Hide();
	 self.slice:ClearAllPoints();

	 self.chip2:Hide();
	 self.chip2:ClearAllPoints();
	 self.slice2:Hide();
	 self.slice2:ClearAllPoints();

	 self.quadExtra:Hide();
	 self.quadExtra:ClearAllPoints();

	 -- Remember this for next time
	 self.lastQuad = quad;
      end

      -- Remember the angle for next time
      self.angle = angle;
      self.startAngle = startAngle;

      -- Get quadrant-specific elements
      local T = self.quadrants[math.mod(quad-1,4)+1];
      local SS = setSubsetFuncs[effQuad];
      local TM = ringTexCoordMapFuncs[effQuad];
      local STM = sliceTexCoordMapFuncs[effQuad];

      if ((quad ~= 5) or (angle == 360)) then
	 self.quadExtra:Hide();
      end

      if (quad == 1) then
	 -- Start and end are in the same quadrant, normal subset of quadrant
	 if (startA == 0) then
	    self.quadExtra:Hide();
	    self.chip2:Hide();
	    self.slice2:Hide();
	 else
	    T:Hide();
	    self:DoQuadrantPartial(startA, endA, T, SS, TM,
				   nil, nil, nil, nil, nil,
				   STM)
	    return;
	 end
      end

      -- Quick hit to handle all-visible case
      if (angle == 360) then
	 endA = 90;
      end
      -- Do END of ring
      self:DoQuadrant(endA, T, SS, TM)

      if (quad ~= 1) then
	 local T = self.quadrants[1];
	 if (quad == 5) then
	    if (angle == 360) then
	       return;
	    end
	    self.quadExtra:SetTexture(T:GetTexture());
	    T = self.quadExtra;
	 end
	 local Qf = math.mod(quadOfs,4)+1;
	 local SS = setSubsetFuncs[Qf];
	 local TM = ringTexCoordMapFuncs[Qf];
	 -- Do START of ring
	 self:DoQuadrantReversed(90-startA, T, SS, TM, self.slice2, self.chip2)
      end
   end

   ---------------------------------------------------------------------------
   -- Some handy method functions


   -- StatsRingRingTemplate:ForceRefresh()
   --
   -- Clear saved state and refresh the display
   --

   function METHODS.ForceRefresh(self)
      local savedAngle = self.angle;
      self.angle = nil;
      self.lastQuad = nil;

      if (savedAngle) then
	 self:SetAngle(savedAngle);
      end
   end

   -- StatsRingRingTemplate:CallTextureMethod(method, ...)
   --
   -- Invokes the named method on all of the textures in the ring,
   -- passing in whatever arguments are given.
   --
   --  e.g. ring:CallTextureMethod("SetVertexColor", 1.0, 0.5, 0.2, 1.0);

   function METHODS.CallTextureMethod(self, method, ...)
      self.quadrants[1][method](self.quadrants[1], unpack(arg));
      self.quadrants[2][method](self.quadrants[2], unpack(arg));
      self.quadrants[3][method](self.quadrants[3], unpack(arg));
      self.quadrants[4][method](self.quadrants[4], unpack(arg));
      self.chip[method](self.chip, unpack(arg));
      self.slice[method](self.slice, unpack(arg));
      if (self.chip2) then
	 self.chip2[method](self.chip2, unpack(arg));
	 self.slice2[method](self.slice2, unpack(arg));
	 self.quadExtra[method](self.quadExtra, unpack(arg));
      end
   end


   -- StatsRingRingTemplate:SetRingTextures(ringFactor,ringTexFile[,sliceTexFile])
   --
   -- Sets the textures to use for this ring
   --
   -- Param:
   --   ringFactor   - The ring factor (Inner Radius / Outer Radius)
   --   ringTexFile  - The ring texture filename
   --   sliceTexFile - The slice texture filename (optional)

   function METHODS.SetRingTextures(self, ringFactor,
				     ringTexture,
				     sliceTexture)
      self.ringFactor = ringFactor;

      for i=1,4 do
	 self.quadrants[i]:SetTexture(ringTexture);
      end
      self.chip:SetTexture(ringTexture);
      if (sliceTexture) then self.slice:SetTexture(sliceTexture) end;
      if (self.chip2) then
	 self.chip2:SetTexture(ringTexture);
	 if (sliceTexture) then self.slice2:SetTexture(sliceTexture) end;
	 self.quadExtra:SetTexture(ringTexture);
      end

      self:ForceRefresh();
   end

   -- Method function to set a quadrant offset (CLOCKWISE, regardless of
   -- direction of growth)
   --
   -- Param:
   --  self  - The ring template instance
   --  offset - The quadrant offset (0-3)
   function METHODS.SetQuadrantOffset(self, offset)
      offset = math.floor(offset);
      if (offset < 0) then
	 offset = math.mod(-offset, 4); 
	 offset = math.mod(4-offset, 4); 
      else
	 offset = math.mod(offset, 4);
      end
      
      if (self.quadOffset == offset) then
	 return
      end

      self.quadOffset = offset;
      self:ForceRefresh();
   end

   -- Method function to set whether or not ring growth is reversed.
   --
   -- Param:
   --  self       - The ring template instance
   --  isReversed - Whether to reverse or not
   function METHODS.SetReversed(self, isReversed)
      if (isReversed) then
	 isReversed = true;
      else
	 isReversed = nil;
      end
      if (isReversed == self.reversed) then
	 return;
      end
      self.reversed = isReversed;
      self:ForceRefresh();
   end

   -- The SetupFrame method, call this for each template object to set it up
   -- and get things going
   --
   -- This is actually a ringInstance method, and not a frame method.
   function INSTANCE.SetupFrame(self, frame)
      -- Grab texture references and frame name
      frame.name = frame:GetName();
      local isPlus = getglobal(frame.name .. "TX1") ~= nil;
      if (isPlus) then
	 frame.isPlus = true;
      end

      local newMeta = {};
      local oldMeta = getmetatable(frame);
      for k,v in pairs(oldMeta) do newMeta[k] = v; end
      frame.ringInstance = self;
      frame.ringMethods = self.methods;

      function newMeta.__index(table, index) 
	 return oldMeta.__index(table,index) or table.ringMethods[index];
      end

      setmetatable(frame, newMeta);

      frame.quadrants = {};
      frame.quadrants[1] = getglobal(frame.name .. "TQ1");
      if (not frame.quadrants[1]) then
	 message("Unable to find first quadrant for '" .. frame.name .. "'");
      end
      frame.quadrants[2] = getglobal(frame.name .. "TQ2");
      frame.quadrants[3] = getglobal(frame.name .. "TQ3");
      frame.quadrants[4] = getglobal(frame.name .. "TQ4");
      frame.chip         = getglobal(frame.name .. "TC1");
      frame.slice        = getglobal(frame.name .. "TS1");
      if (isPlus) then
	 frame.chip2        = getglobal(frame.name .. "TC2");
	 frame.slice2       = getglobal(frame.name .. "TS2");
	 frame.quadExtra    = getglobal(frame.name .. "TX1");
      end

      -- Initialize size and default texture ringFactor
      frame.radius = frame:GetWidth() / 2;
      frame.ringFactor = 0.9;
      
      frame.quadOffset = 0;

      -- Set angle to zero (initializes texture visibility)
      frame:SetAngle(0);
   end

   INSTANCE.subVersion = RING_TEMPLATE_SUB_VERSION;

   return INSTANCE;
end

-- Check to see if there's a version already installed
local tmpl = IrielRingTemplate;
if (tmpl == nil) then
   tmpl = {};
   tmpl.instances = {};
   tmpl.bestVersion = 0;
end

-- Check to see if this is the most recent version
local newSub = nil;
local inst = tmpl.instances[RING_TEMPLATE_COMPATIBILITY];

if ((not inst) or (inst.subVersion < RING_TEMPLATE_SUB_VERSION)) then
   inst = GetTemplateInstance();
   newSub = 1;
   tmpl.instances[RING_TEMPLATE_COMPATIBILITY] = inst;
end

-- If this is a new better version, then replace the old manager
if (newSub and (RING_TEMPLATE_COMPATIBILITY > tmpl.bestVersion)) then
   tmpl.bestVersion = RING_TEMPLATE_COMPATIBILITY;

   function tmpl:GetInstance(compatVersion) 
      return self.instances[compatVersion or self.bestVersion];
   end

   function tmpl:InitFrame(frame, compatVersion)
      self:GetInstance(compatVersion):SetupFrame(frame);
   end
end

if (not IrielRingTemplate) then
   IrielRingTemplate = tmpl;
end

-- Allow create function to be garbage collected.
GetTemplateInstance = nil;
