-- Code to track mouseovers between Sliders (and potentially other things)
-- Primarily for preview purposes in StatRings but it could have other
-- uses.

local function DEBUG(msg)
   DEFAULT_CHAT_FRAME:AddMessage("SROMT: " .. msg);
end

-- Map of <Frame> to state flag indicating current mouseover state of
-- a component.
local MOUSEOVER_STATE = {};
local MOUSEOVER_STATE_MOUSEOVER = "mouseover";
local MOUSEOVER_STATE_MOUSEDOWN = "mousedown";

local MOUSEOVER_UPDATE_PERIOD = 0.2

-- Current 'active' mouseover
local MOUSEOVER_OWNER;
-- How long before the next OnUpdate check
local MOUSEOVER_UPDATE_TIMER = 0;

-- Claim mouseover for a new frame, returns 1 if the new frame is
-- eligible to do so, or nil if it must pass.
local function MouseoverTrackClaimOwnership(frame)
   if (not frame) then
      return;
   end
   --DEBUG("ClaimOwnership(" .. frame:GetName() .. ")");
   local oldOwner = MOUSEOVER_OWNER;
   local lossFunc;
   if (oldOwner) then
      if (oldOwner == frame) then
	 -- Strange case but we should check for it
	 --DEBUG("  Accept (already owner)");
	 return 1;
      end
      local oldState = MOUSEOVER_STATE[oldOwner];
      if (oldState == MOUSEOVER_STATE_MOUSEDOWN) then
	 --DEBUG("  Reject (Old mouse down)");
	 return;
      end
      lossFunc = oldOwner.OnMouseoverLoss;
   end
   MOUSEOVER_OWNER = frame;

   local gainFunc = frame.OnMouseoverGain;

   if (gainFunc) then
      gainFunc(frame, (oldOwner and true));
   end
   if (lossFunc) then
      lossFunc(oldOwner, true);
   end
   --DEBUG("  Accept");
   return 1;
end

-- Release mouseover ownership
local function MouseoverTrackReleaseOwnership(frame)
   if ((not frame) or (MOUSEOVER_OWNER ~= frame)) then
      return;
   end
   local oldOwner = MOUSEOVER_OWNER;
   MOUSEOVER_OWNER = nil;
   local lossFunc = oldOwner.OnMouseoverLoss;
   if (lossFunc) then
      lossFunc(oldOwner, false);
   end
end

local MOT_Slider_MouseIsOver;

local MOT_Slider_OnEnter;
local MOT_Slider_OnUpdate;
local MOT_Slider_OnMouseDown;
local MOT_Slider_OnMouseUp;
local MOT_Slider_OnHide;

-- Test to see if the mouse is over this slider
-- TODO - Add support for vertical slider
function MOT_Slider_MouseIsOver(frame)
   return MouseIsOver(frame, 11, -11, -1, 1);
end

-- OnEnter handler - If the mouse isn't yet over this slider, try and
-- obtain ownership of the mouseover.
function MOT_Slider_OnEnter()
   local this = this;
   local mo = MOUSEOVER_STATE[this];
   if (not mo) then
      if (MouseoverTrackClaimOwnership(this)) then
	 MOUSEOVER_STATE[this] = MOUSEOVER_STATE_MOUSEOVER;
	 MOUSEOVER_UPDATE_TIMER = MOUSEOVER_UPDATE_PERIOD; -- Refresh time
	 this:SetScript("OnUpdate", MOT_Slider_OnUpdate);
	 this:SetScript("OnHide",   MOT_Slider_OnHide);
      end
   end
end

-- OnUpdate handler - Ugly but necessary, periodically check to see
-- if the mouse has left the frame yet.
function MOT_Slider_OnUpdate()
   local this = this;
   local mo = MOUSEOVER_STATE[this];
   -- If we dont think we're moused over, then leave
   if (not mo) then
      this:SetScript("OnUpdate", nil);
      return;
   end
   -- If we no longer own the mouseover claim, then leave
   if (this ~= MOUSEOVER_OWNER) then
      MOUSEOVER_STATE[this] = nil;
      this:SetScript("OnUpdate", nil);
      return;
   end
   
   MOUSEOVER_UPDATE_TIMER = MOUSEOVER_UPDATE_TIMER - arg1;
   if (MOUSEOVER_UPDATE_TIMER > 0) then 
      return;
   end
   if ((mo == MOUSEOVER_STATE_MOUSEDOWN)
       or MOT_Slider_MouseIsOver(this)) then
      MOUSEOVER_UPDATE_TIMER = MOUSEOVER_UPDATE_PERIOD; -- Refresh time
      return;
   end
   -- Mouse has left!
   --DEBUG("Mouse has left " .. this:GetName()); 
   MOUSEOVER_STATE[this] = nil;
   this:SetScript("OnUpdate",  nil);
   this:SetScript("OnHide",    nil);
   this:SetScript("OnMouseUp", nil);
   MouseoverTrackReleaseOwnership(this);
end

-- OnMouseDown handler - If we have mouseover already then just set
-- the down flag, otherwise try and obtain it, and setup handlers
-- accordingly
function MOT_Slider_OnMouseDown()
   local this = this;
   local mo = MOUSEOVER_STATE[this];
   if (mo) then
      if (mo ~= MOUSEOVER_STATE_MOUSEDOWN) then
	 MOUSEOVER_STATE[this] = MOUSEOVER_STATE_MOUSEDOWN;
	 this:SetScript("OnMouseUp", MOT_Slider_OnMouseUp);
      end
   elseif (MOT_Slider_MouseIsOver(this)) then
      if (MouseoverTrackClaimOwnership(this)) then
	 MOUSEOVER_STATE[this] = MOUSEOVER_STATE_MOUSEDOWN;
	 MOUSEOVER_UPDATE_TIMER = MOUSEOVER_UPDATE_PERIOD; -- Refresh time
	 this:SetScript("OnUpdate",  MOT_Slider_OnUpdate);
	 this:SetScript("OnHide",    MOT_Slider_OnHide);
	 this:SetScript("OnMouseUp", MOT_Slider_OnMouseUp);
      end
   end
end

-- OnMouseUpHandler - If the mouse is still inside the object
-- then simply revert the state, otherwise release the mouseover
function MOT_Slider_OnMouseUp()
   local this = this;
   local mo = MOUSEOVER_STATE[this];
   if (mo and (mo == MOUSEOVER_STATE_MOUSEDOWN)) then
      this:SetScript("OnMouseUp", nil);
      if (this == MOUSEOVER_OWNER) then
	 if (MOT_Slider_MouseIsOver(this)) then
	    MOUSEOVER_STATE[this] = MOUSEOVER_STATE_MOUSEOVER;
	    MOUSEOVER_UPDATE_TIMER = MOUSEOVER_UPDATE_PERIOD; -- Refresh time
	    return;
	 end
      end

      MOUSEOVER_STATE[this] = nil;
      this:SetScript("OnUpdate",  nil);
      this:SetScript("OnHide",    nil);
      this:SetScript("OnMouseUp", nil);
      MouseoverTrackReleaseOwnership(this);
   end
end

-- OnHide handler - Automatically relinquish mouseoverness.
function MOT_Slider_OnHide()
   local this = this;
   local mo = MOUSEOVER_STATE[this];
   -- Mouse has left!
   if (mo) then
      --DEBUG("Mouse has left " .. this:GetName() .. " (Hide)"); 
      this:SetScript("OnUpdate", nil);
      this:SetScript("OnHide",   nil);
      if (mo == MOUSEOVER_STATE_MOUSEDOWN) then
	 this:SetScript("OnMouseUp", nil);
      end
      MOUSEOVER_STATE[this] = nil;
      MouseoverTrackReleaseOwnership(this);
   end
end

-- Call this on a slider to make it eligible for mouseover tracking.
function StatRingsOption_MOT_SetupSlider(frame)
   frame:SetScript("OnEnter",     MOT_Slider_OnEnter);
   frame:SetScript("OnMouseDown", MOT_Slider_OnMouseDown);
end
