-- Scaffolding for ring structure, replaces StatRingsLayout

-- Metatable of methods for scaffold objects
local Scaffold = {};
-- Metatable of methods for scaffold compiler objects
local ScaffoldCompiler = {};

-- We'll be calling these a lot, so grab local versions (Yes, some of them
-- are defined in compat.lua but I can't count on that can I?)
local rad, deg = math.rad, math.deg;
local sin, cos, atan2 = math.sin, math.cos, math.atan2;
local sqrt,abs,mod = math.sqrt, math.abs, math.mod;

local function SinCosDeg(deg)
   local r = rad(deg);
   return sin(r), cos(r);
end


---------------------------------------------------------------------------
-- UTILITY FUNCTIONS

-- Output debug message
local function DEBUG(msg)
   DEFAULT_CHAT_FRAME:AddMessage("[Scaffold] " .. msg);
end

-- This is handy in a few places, but must be kept clean!
local EMPTY_TABLE = {};
setmetatable(EMPTY_TABLE, { 
		__newindex = function(t,k,v) 
				DEBUG("Attempt to add key " 
				      .. k .. " to EMPTY_TABLE");
			     end; 
	     });

-- Copy group settings into a group structure, returns true if anything changed
local function ScaffoldUtil_CopyGroupSettings(group, groupSpec, groupOverride)
   local changed = nil;

   -- Grab appropriate values (all independent)
   local size   = math.floor(tonumber(groupOverride.size) 
			     or tonumber(groupSpec.size) or 0);
   if (size < 0) then size = 0; end
   local spread = groupOverride.spread or groupSpec.spread or "none";
   local weight = (tonumber(groupOverride.weight) 
		   or tonumber(groupSpec.weight) or 0);
   local spinReverse = false;
   if (groupOverride.spinReverse ~= nil) then
      spinReverse = (groupOverride.spinReverse and true) or false;
   elseif (groupSpec.spinReverse ~= nil) then
      spinReverse = (groupSpec.spinReverse and true) or false;
   end

   -- Compare with current settings
   if ((size ~= group.size)
       or (spread ~= group.spread) 
	  or (weight ~= group.weight)
	  or (spinReverse ~= group.spinReverse)) then
      group.size = size;
      group.spread = spread;
      group.weight = weight;
      group.spinReverse = spinReverse;
      if (size == 0) then
	 group.effSize = table.getn(group.members);
      else
	 group.effSize = size;
      end
      return true;
   end
end

local function NormalizeAnchor(src, dest)
   if (type(src) ~= "table") then
      src = EMPTY_TABLE;
   end
   if (not dest) then
      dest = {};
   end
   dest.angle = tonumber(src.angle) or 0;
   dest.length = tonumber(src.length) or 0;
   dest.scale = tonumber(src.scale) or 1;
   dest.spin = tonumber(src.spin) or 0;
   dest.relSpin = (src.relSpin and true) or false;
   return dest;
end

local function DescribeNode(node)
   local t = node.type;
   if (t == "root") then
      return "Root";
   elseif (t == "group") then
      return "Group";
   elseif (t == "member") then
      return "Member of '" .. node.parent .. "'";
   elseif (t == "inert") then
      return "Inactive Node";
   elseif (t == "node") then
      return "Node";
   elseif (not t) then
      return "<nil> Node";
   end
      
   return "<" .. t .. "> Node";
end

local function MarkChildrenDirty(self, nodes, node)
   local anyDirty;
   for _,n in ipairs(node.children) do
      local child = nodes[n];
      if (not child.dirty) then
	 child.dirty = true;
	 anyDirty = true;
      end
   end
   if (anyDirty) then
      self.dirty = true;
   end
end

-- Compare two locations, and copy the base location to the copy if
-- they dont match.
-- Returns true if they didn't match, false if they did
local function CompareCopyLoc(baseLoc, copyLoc, copyObj, locName, copyLocName)
   if (baseLoc == copyLoc) then
      return false;
   end
   if (not baseLoc) then
      copyObj[locName] = nil;
      return true;
   end
   if (copyLoc) then
      local anyChange;
      for i=1,4 do
	 if (baseLoc[i] ~= copyLoc[i]) then
	    anyChange = true;
	    break;
	 end
      end
      if (not anyChange) then
	 return false;
      end
   else
      if (not copyLocName) then copyLocName = "_" .. locName; end
      copyLoc = copyObj[copyLocName];
      if (not copyLoc) then
	 copyLoc = {0,0,1,0};
	 copyObj[copyLocName] = copyLoc;
      end
      copyObj[locName] = copyLoc;
   end
   for i=1,4 do
      copyLoc[i] = baseLoc[i];
   end
   return true;
end

local function CopyAnchor(src, dest)
   if (not src) then
      return nil;
   elseif (not dest) then
      dest = {};
   end

   dest.inactive = src.inactive;
   dest.angle    = src.angle;
   dest.length   = src.length;
   dest.scale    = src.scale
   dest.spin     = src.spin;
   dest.relSpin  = src.relSpin;
end

local function CopyGroupSpec(src, dest)
   if (not src) then
      return nil;
   elseif (not dest) then
      dest = {};
   end

   dest.size        = src.size;
   dest.spread      = src.spread;
   dest.weight      = src.weight;
   dest.spinReverse = src.spinReverse;
end

local function ApplyAnchor(anchor, pX, pY, pScale, pSpin)
   if ((not anchor) or (anchor.inactive)) then
      return;
   end
   local scale = anchor.scale * pScale;
   local angle = anchor.angle + pSpin;
   local len   = anchor.length;
		
   if (len > 0) then
      local s,c = SinCosDeg(angle);
      if (len > 1) then
	 len = pScale + (len - 1) * scale;
      else
	 len = len * pScale;
      end
      pX = pX + s*len;
      pY = pY + c*len;
   end

   local spin = anchor.spin;
   if (anchor.relSpin) then
      spin = spin + angle;
   end
   return pX, pY, scale, spin;
end

---------------------------------------------------------------------------
-- SCAFFOLD COMPILATION

local ScaffoldCompiler_Meta = { __index = ScaffoldCompiler; }

-- Add an error to the compiled result
function ScaffoldCompiler:AddError(msg)
   local errs = self.errors;
   if (not errs) then
      errs = {};
      self.errors = errs;
   end
   table.insert(errs, msg);
   DEBUG("ERROR: " .. msg);
end

-- Process creation of a normal node (Assumes that name is unused)
--
-- name - The name of the node to create
-- spec - The node's specification structure
-- overSpec - The node's override structure (must exist, can be empty)
--            MUST NOT BE EDITED
function ScaffoldCompiler:CreateNormalNode(name, spec, overSpec)
   local nodes = self.nodes;
   local node = { 
      type = "node";
      dirty = true;
      children = {};
      parent = overSpec.parent or spec.parent or "root";
      defaultAnchor = NormalizeAnchor(spec.anchor);
   };
   -- Now get parameters from specification
   local followType, followNode;
   if (overSpec.followType) then
      followType, followNode = overSpec.followType,
      (overSpec.followNode or "root");
   elseif (spec.followType) then
      followType, followNode = spec.followType, spec.followNode or "root";
   else
      followType = "none";
   end

   -- Pick the appropriate follow type
   if ((followType == "anchor") or (followType == "node")) then
      node.followType = followType;
      node.followNode = followNode;
   else
      node.followType = followType;
      node.followNode = nil;
   end

   if (followType == "node") then
      -- Create storage arrays for followed node and parent
      node._nLoc = {0,0,1,0}
      node._pLoc = {0,0,1,0}
   end
   
   nodes[name] = node;

   --DEBUG("Created normal node " .. name);
end

-- Process creation of a group node (Assumes that name is unused), and
-- of all member nodes (Will do collision detection on names)
--
-- name - The name of the group to create
-- spec - The group's specification structure
-- overSpec - The group's override structure (must exist, can be empty)
--            MUST NOT BE EDITED
function ScaffoldCompiler:CreateGroupNode(name, spec, overSpec)
   local nodes = self.nodes;
   local startNode = name .. "Start";
   local group = { 
      type = "group";
      dirty = true;
      _sLoc = {0,0,1,0};
      _eLoc = {0,0,1,0};
      children = {};
      members = {};
      startNode = name .. "Start";
      endNode = name .. "End";
      reAnchor = true;
   };

   -- Add to nodes now so we get better errors with members
   nodes[name] = group;

   -- Now do members
   for _, memberName in ipairs(spec.members) do
      if (nodes[memberName]) then
	 self:AddError("Ignoring duplicate node '" .. memberName
		       .. "' as member of '" .. name .. "'; Already seen as " 
			  .. DescribeNode(nodes[nodeName]));
      else
	 local member = {
	    type = "member";
	    parent = name;
	    dirty = true;
	    active = false;
	    children = {};
	    index = table.getn(group.members) + 1;
	 };
	 table.insert(group.members, memberName);
	 nodes[memberName] = member;
	 --DEBUG("Created member node " .. memberName);
      end
   end
   
   -- And finally copy settings
   ScaffoldUtil_CopyGroupSettings(group, spec, overSpec);

   --DEBUG("Created group node " .. name);
end

-- Create the scaffold nodes
function ScaffoldCompiler:CreateNodes(overrides)
   local spec = self.spec;
   local nodes = self.nodes;
   if (not overrides) then
      overrides = EMPTY_TABLE;
   end
   
   -- Start with root nodes, they cannot be overriden
   if (spec.roots) then
      for rootName,v in pairs(spec.roots) do
	 local rootLoc = {
	    tonumber(v[1]) or 0; -- X
	    tonumber(v[2]) or 0; -- Y 
	    tonumber(v[3]) or 1; -- Scale
	    tonumber(v[4]) or 0; -- Spin
	 };
	 nodes[rootName] = {
	    type = "root";
	    loc = rootLoc;
	    _loc = rootLoc;
	    children = {};
	 };
	 --DEBUG("Created root node " .. rootName);
      end
   end
   -- Create the "root" root if it's not already declared
   if (not nodes["root"]) then
      local rootLoc = {0,0,1,0};
      nodes["root"] = {
	 type = "root";
	 loc = rootLoc;
	 _loc = rootLoc;
	 children = {};
      };
      --DEBUG("Created default root node");
   end

   -- Now make the nodes themselves
   if (spec.nodes) then
      for nodeName, v in pairs(spec.nodes) do
	 if (nodes[nodeName]) then
	    self:AddError("Ignoring duplicate node '" .. nodeName
			  .. "'; Already seen as " 
			     .. DescribeNode(nodes[nodeName]));
	 else
	    -- Work out what kind of node this is
	    if (type(v.members) == "table") then
	       self:CreateGroupNode(nodeName, v, 
				    overrides[nodeName] or EMPTY_TABLE);
	    else
	       self:CreateNormalNode(nodeName, v, 
				     overrides[nodeName] or EMPTY_TABLE);
	    end
	 end
      end
   end
end

-- Do the dependencies for one node
--
--  nodeName - The name of the node node to process
--  node     - The node itself
function ScaffoldCompiler:DoNodeDependencies(nodeName, node)
   local nodes = self.nodes;
   
   -- If this node has a location field then it's already done
   if (node._loc) then
      return;
   end
   --DEBUG("  DoNodeDependencies(" .. nodeName .. ") active");
   
   node.inProcess = true;
   
   local depa, depb;
   local nodeType = node.type;
   local broken = node.broken;
   if (nodeType == "node") then
      depa = node.parent;
      depb = node.followNode;
   elseif (nodeType == "member") then
      depa = node.parent;
   elseif (nodeType == "root") then
      -- No dependencies
   elseif (nodeType == "group") then
      depa = node.startNode;
      depb = node.endNode;
   else
      self:AddError("Unknown node '" .. nodeName
		    .. "' (" .. DescribeNode(node) 
		       .. ") when making dependencies");
      broken = true;
   end

   local nodeA, nodeB;
   if (depa) then
      --DEBUG("    Exploring dependency " .. depa .. " for " .. nodeName);
      nodeA = nodes[depa];
      if (depa == nodeName) then
	 self:AddError("Node '" .. nodeName .. "' ("
		       .. DescribeNode(node) .. ") cannot depend on itself!");
	 broken = true;
      elseif (not nodeA) then
	 self:AddError("Node '" .. nodeName .. "' ("
		       .. DescribeNode(node) .. ") depends on missig node '"
		       .. depa .. "'");
	 broken = true;
      end
   end

   if (depb) then
      --DEBUG("    Exploring dependency " .. depb .. " for " .. nodeName);
      nodeB = nodes[depb];
      if (depa == depb) then
	 -- Already processed this one
      elseif (depb == nodeName) then
	 self:AddError("Node '" .. nodeName .. "' ("
		       .. DescribeNode(node) .. ") cannot depend on itself!");
	 broken = true;
      elseif (not nodeB) then
	 self:AddError("Node '" .. nodeName .. "' ("
		       .. DescribeNode(node) .. ") depends on missig node '"
		       .. depb .. "'");
	 broken = true;
      end
   end

   if ((not broken) and nodeA and (not nodeA._loc)) then
      if (nodeA.inProcess) then
	 self:AddError("Node '" .. nodeName .. "' has dependency loop to '"
		       .. depa .. "'");
	 broken = true;
      else
	 self:DoNodeDependencies(depa, nodeA);
      end
   end
   
   if ((not broken) and nodeB and (not nodeB._loc)) then
      if (nodeB.inProcess) then
	 self:AddError("Node '" .. nodeName .. "' has dependency loop to '"
		       .. depb .. "'");
	 broken = true;
      else
	 self:DoNodeDependencies(depb, nodeB);
      end
   end
   
   if ((not broken) and ((nodeType == "node") 
			 or (node.followType == "anchor")) and nodeB) then
      if (nodeB.type == "root") then
	 self:AddError("Node '" .. nodeName .. "' cannot use anchor from"
		       .. " root node '" .. depb .. "'");
	 broken = true;
      end
   end

   if ((nodeA and nodeA.broken) or (nodeB and nodeB.broken)) then
      broken = true;
   end

   node._loc = {0,0,1,0};
   node.inProcess = nil;
   
   if (broken) then
      node.broken = true;
      return;
   end

   if (nodeA) then
      table.insert(nodeA.children, nodeName);
   end
   if (nodeB and (nodeA ~= nodeB)) then
      table.insert(nodeB.children, nodeName);
   end
   table.insert(self.scan, nodeName);
end

-- Scan the dependencies of the created nodes, and populate broken and
-- children lists
function ScaffoldCompiler:ScanDependencies()
   if (self.scan[1]) then
      DEBUG("Cannot call ScanDependencies again!");
      return;
   end
   -- Add root nodes to scan, since they have locations
   for nodeName, node in pairs(self.nodes) do
      if (node.type == "root") then
	 table.insert(self.scan, nodeName);
      end
   end
   -- Scan all nodes
   for nodeName, node in pairs(self.nodes) do
      --DEBUG("Doing node dependency for " .. nodeName);
      self:DoNodeDependencies(nodeName, node);
   end
   -- And add broken nodes.
   for nodeName, node in pairs(self.nodes) do
      if (node.broken) then
	 table.insert(scan, nodeName);
      end
   end
end

-- Determine the default anchor for the specified node
-- nodeName - the name of the node being processed
-- node - The node itself
function ScaffoldCompiler:DoDefaultAnchor(nodeName, node)
   local nodeType = node.type;
   local anchors = self.anchors;

   if (nodeType == "root") then
      -- Roots have no anchor
      return;
   elseif (nodeType == "member") then
      -- Members have a derived anchor from the parent group
      local anchor = {
	 inactive = true; -- For now it's inactive
	 derived = node.parent; -- Which node owns controlling this
      };
      anchors[nodeName] = anchor;
      return;
   elseif (nodeType == "group") then
      -- Groups have a derived anchor from themselves, possibly inactive
      -- if the group is broken.
      -- Also use this step as an opportunity to set the group's parent.

      local anchor;
      if (node.broken) then
	 anchor = {
	    inactive = true; -- For now it's inactive
	    derived = nodeName; -- Which node owns controlling this
	 };
      else
	 local nodes = self.nodes;
	 local startNode = nodes[node.startNode];
	 local endNode = nodes[node.endNode];

	 if (startNode.parent == endNode.parent) then
	    node.parent = startNode.parent;
	 else
	    node.parent = node.startNode;
	 end

	 anchor = {
	    derived = nodeName;
	    angle = 0;
	    length = 0;
	    scale = 1;
	    spin = 0;
	    relSpin = false;
	 };
      end
      anchors[nodeName] = anchor;
      return;
   elseif (nodeType == "node") then
      -- Normal nodes have either an explicit anchor, or if they have a
      -- follow type, they use a derived anchor from themselves.
      local followType = node.followType;
      local anchor;

      if ((followType == "node") or (followType == "anchor")) then
	 anchor = {
	    inactive = true;
	    derived = nodeName;
	 };
      else
	 anchor = {};
	 for k,v in pairs(node.defaultAnchor) do
	    anchor[k] = v;
	 end
      end
      anchors[nodeName] = anchor;
      return;
   end
      
   self:AddError("Cannot derive anchor for '" .. nodeName .. "' ("
		 .. DescribeNode(node) .. "'");
end

function ScaffoldCompiler:DefaultAnchors()
   for nodeName, node in pairs(self.nodes) do
      self:DoDefaultAnchor(nodeName, node);
   end
end

function ScaffoldCompiler:GetNumErrors()
   local errs = self.errors;
   return (errs and table.getn(errs)) or 0;
end

function ScaffoldCompiler:GetError(idx)
   local errs = self.errors;
   return (errs and errs[idx]);
end

function ScaffoldCompiler:GetErrors()
   return self.errors;
end

function ScaffoldCompiler:Compile(overrideSpec)
   self:CreateNodes(overrideSpec);
   self:ScanDependencies();
   self:DefaultAnchors();
end

function StatRingsScaffoldCompiler_New(spec)
   local data = {
      spec = spec,
      nodes = { },
      anchors = { },
      scan = { },
   };
   setmetatable(data, ScaffoldCompiler_Meta);

   return data;
end

function StatRingsScaffoldCompiler_Run(spec, overrideSpec)
   local compiler = StatRingsScaffoldCompiler_New(spec);
   compiler:Compile(overrideSpec);
   return compiler;
end

---------------------------------------------------------------------------

function Scaffold:Error(msg)
   _ERRORMESSAGE("Scaffold: " .. msg);
end

-- ifApplied,errors = Scaffold:Compile(spec, overrideSpec, useWithErrors)
--   Parameters:
--     spec -- The scaffold specification structure
--     overrideSpec -- Node override specifications (keyed by node)
--     useWithErrors -- Flag, set to true if the compilation should be accepted
--                      even if the spec has errors. Default false.
--
--   Returns
--     ifApplied -- Flag - True if the spec was applied, else false
--     errorArray -- Array - Array of errors if any generated, else nil
--
--     Compiles the provided specification (With setting overrides from
--     overrideSpec, if provided) and then, if successful or useWithErrors is
--     set, apply the result to this scaffold.
--
--     You should call Scaffold:Update() if the spec was applied.
function Scaffold:Compile(spec, overrideSpec, useWithErrors) 
   local compiler = StatRingsScaffoldCompiler_Run(spec, overrideSpec);
   local compileErrors = compiler:GetErrors();
   -- Check to see if compilation failed
   if (compileErrors and (not useWithErrors)) then
      return nil, compileErrors;
   end

   -- Grab references to things
   local oldNodes = self.nodes;
   local inertNodes = self.inertNodes;
   local nodes = compiler.nodes;

   local oldAnchors = self.anchors;
   local anchors = compiler.anchors;

   local visibilities = self.visibilityManager.states or EMPTY_TABLE;

   -- Drop new ones in over old ones
   self.scan = compiler.scan;
   self.nodes = nodes;
   self.anchors = anchors;

   local changes = false;

   -- And now process the differences for nodes...
   for name, node in pairs(nodes) do
      local oldNode = oldNodes[node];
      if (oldNode) then
	 oldNodes[node] = nil;
      else 
	 oldNode = inertNodes[node];
	 if (oldNode) then
	    inertNodes[node] = nil;
	 end
      end

      local nodeType = node.type;
      if (oldNode) then
	 local oldType = oldNode.type;

	 -- First, copy over event information if it exists.
	 node.attached = oldNode.attached;
	 node.dispatching = oldNode.dispatching;
	 node.changed = oldNode.changed;

	 -- Check if type is the same
	 if (nodeType ~= oldType) then
	    node.changed = true;
	 elseif (nodeType == "root") then
	    -- If the old root location was overriden, copy it
	    if (oldNode.override) then
	       node.override = true;
	       node.loc = oldNode.loc;
	       node._loc = oldNode._loc;
	    end
	 end

	 if (nodeType == "root") then
	    -- For root, cannot always copy location since it's fixed
	    local oldLoc = oldNode.loc;
	    local loc = node.loc;

	    if (loc and oldLoc) then
	       if (loc ~= oldLoc) then
		  local locDiff;
		  for i=1,4 do 
		     if (loc[i] ~= oldLoc[i]) then
			locDiff = true;
			break;
		     end
		  end
		  if (locDiff) then
		     node.changed = true;
		  end
	       end
	    elseif (loc ~= oldLoc) then
	       node.changed = true;
	    end
	 else
	    -- Location is derived, so copy old one in order to avoid
	    -- firing a change if one isn't needed
	    if (node.broken) then
	       if (oldNode.loc) then
		  node.changed = true;
	       end
	    else
	       node.loc = oldNode.loc;
	       node._loc = oldNode._loc;
	    end
	 end
      end
      local nVisState = visibilities[name]
	 or (nodeType == "member" and "HIDDEN") or "SHOWN";
      if (node.visState ~= nVisState) then
	 node.visState = nVisState;
	 node.changed = 1;
      end
      if (node.changed) then changes = true; end
   end
   
   -- Next make any former nodes inert
   for name, oldNode in pairs(oldNodes) do
      local inert = {
	 type = "inert";
	 attached = oldNode.attached;
	 dispatching = oldNode.dispatching;
	 changed = oldNode.changed;
	 visState = "HIDDEN";
      }
      if ((oldNode.loc) or (oldNode.visState ~= "HIDDEN")) then
	 -- If there used to be a location, now there isn't, fire event
	 inert.changed = 1;
      end
      inertNodes[name] = inert;
      if (inert.changed) then 
	 self.inertChanged = true;
	 changes = true; 
      end
   end

   -- Clean up inert list
   for name,node in pairs(inertNodes) do
      local att = node.attached;
      if (not (node.dispatching or (att and next(att)))) then
	 insertNodes[name] = nil;
      end
   end
   

   -- Now we're done with nodes, onto anchors
   for name, anchor in pairs(anchors) do
      local oldAnchor = oldAnchors[name];
      -- For anchors that did exist before, and which aren't derived now,
      -- and which were user placed before...
      
      if (oldAnchor and (not anchor.derived) and (oldAnchor.override)) then
	 CopyAnchor(oldAnchor, anchor);
	 anchor.override = true;
      end
   end

   self.changes = changes or self.inertChanged;
   self.dirty = true;

   return true, compileErrors;
end

-- Scaffold:Dump(showAnchors)
--   Parameters:
--     showAnchors -- Flag, set to true if you want node anchors shown
--
-- Outputs the scaffold structure for debugging/morbid curiosity.
function Scaffold:Dump(showAnchors)
   local nodes = self.nodes;
   local anchors = self.anchors;

   for _,name in ipairs(self.scan) do
      local node = nodes[name];
      local label = name .. ": " .. DescribeNode(node) 
	 .. " [" .. node.visState .. "]";
      if (node.loc) then
	 label = label .. " (" .. node.loc[1] .. "," .. node.loc[2] .. ") x"
	    .. node.loc[3] .. " " .. node.loc[4] .. " deg.";
      else
	 label = label .. " <inactive>";
      end
      if (node.dirty) then
	 label = label .. " **DIRTY**";
      end
      if (node.broken) then
	 label = label .. " !!BROKEN!!";
      end
      DEBUG(label);
      if (table.getn(node.children) > 0) then
	 DEBUG("  Children: " .. table.concat(node.children, ", "));
      end
      local anchor = anchors[name];
      if (showAnchors and anchor) then
	 local label = "  Anchor:";
	 if (anchor.derived) then
	    label = label .. " (Derived from " .. anchor.derived .. ")";
	 end
	 if (anchor.inactive) then
	    label = label .. " Inactive";
	 else
	    label = (label 
		     .. " angle=" .. anchor.angle
			.. " length=" .. anchor.length
			.. " scale=" .. anchor.scale
			.. " spin=" .. anchor.spin);
	 end
	 DEBUG(label);
      end
   end
end

-- Scaffold:_UpdateGroupAnchors(name, node)
--   Parameters:
--     name -- The name of the group node
--     group -- The group node itself
--
--   Updates the anchors for the specified group node and its members
function Scaffold:_UpdateGroupAnchors(name, node)
   local nodes = self.nodes;
   local sLoc = node.sLoc;
   local eLoc = node.eLoc;
   local pLoc = node.loc;

   -- Clear re-anchor flag
   node.reAnchor = nil;

   -- Get (or re-calculate) active child count
   local activeCount = node.activeCount;
   local anchors = self.anchors;
   
   if (not activeCount) then
      activeCount = 0;
      for _,n in ipairs(node.members) do
	 local child = nodes[n];
	 if (child.visState ~= "HIDDEN") then
	    activeCount = activeCount + 1;
	 else
	    local anchor = anchors[n];
	    if (not anchor.inactive) then
	       if (not child.dirty) then
		  child.dirty = true;
		  self.dirty = true;
	       end
	       anchor.inactive = true;
	    end
	 end
      end
      node.activeCount = activeCount;
   end

   if (not sLoc and eLoc and pLoc) then
      -- Must invalidate any member anchors that are currently active
      if (activeCount > 0) then
	 for _,n in ipairs(node.members) do
	    local child = nodes[n];
	    if (child.visState ~= "HIDDEN") then
	       local anchor = anchors[n];
	       if (not anchor.inactive) then
		  anchor.inactive = true;
		  if (not child.dirty) then
		     child.dirty = true;
		     self.dirty = true;
		  end
	       end
	    end
	 end
      end
      -- Dont need to go any further
      return;
   end

   if (activeCount > 0) then
      local dAngle,dLength,dScale,dSpin,step,offset;
      local cAngle,cLength,cScale,cSpin,eLength;

      local cX, cY, eAngle, eSpin;

      cX, cY, cScale, cSpin = unpack(sLoc);
      local eX, eY, eScale, eSpin = unpack(eLoc);
      local pX, pY, pScale, pSpin = unpack(pLoc);
      
      local dx,dy = cX-pX, cY-pY;
      cAngle = deg(atan2(dx,dy));
      cLength = sqrt(dx*dx+dy*dy) / pScale;
      
      dx,dy = eX-pX, eY-pY;
      eAngle = deg(atan2(dx,dy));
      eLength = sqrt(dx*dx+dy*dy) / pScale;

      cScale = cScale / pScale;
      dScale = (eScale/pScale - cScale);
      
      dLength = eLength - cLength;
      dAngle = mod(eAngle - cAngle, 360);
      if (dAngle <= -180) then
	 dAngle = dAngle + 360;
      elseif (dAngle > 180) then      
	 dAngle = dAngle - 360;
      end
      if (node.spinReverse) then
	 dAngle = 360 - dAngle;
      end
      
      dSpin = mod(eSpin - cSpin, 360);
      if (dSpin <= -180) then
	 dSpin = dSpin + 360;
      elseif (dSpin > 180) then
	 dSpin = dSpin - 360;
      end
      if (node.spinReverse) then
	 dSpin = -(360 - dSpin);
      end

      cAngle = cAngle - pSpin; -- Subtract parent spin, since it gets re-added
      
      local n = activeCount;
      local nSize = node.effSize;
      if (n > nSize) then n = nSize; end

      offset = 0;
      local spread = node.spread;
      if (spread == "open") then
	 step = 1 / n;
      elseif (spread == "fill") then
	 if (n == 1) then
	    step = 1;
	 else
	    step = 1 / (n - 1);
	 end
      else
	 if (nSize <= 1) then
	    step = 1;
	 else
	    step = 1 / (nSize - 1);
	 end
      end

      local fullwid = (activeCount - 1) * step;
      offset = (1 - fullwid) * (node.weight or 0.5);

      -- Apply values to active children
      local cnt = 0;
      for _,n in ipairs(node.members) do
	 local child = nodes[n];
	 if (child.visState ~= "HIDDEN") then
	    local childAnchor = anchors[n];

	    local amt = offset + step * cnt;
	    cnt = cnt + 1;
	    local nAngle  = cAngle + dAngle * amt;
	    local nLength = mod(cLength + dLength * amt, 360);
	    local nScale  = cScale + dScale * amt;
	    local nSpin   = mod(cSpin + dSpin * amt, 360);
	    if (nSpin < 0) then
	       nSpin = nSpin + 360;
	    end
	    if (nAngle < 0) then
	       nAngle = nAngle + 360;
	    end
	    
	    if (nLength > 1) then
	       nLength = 1 + (nLength - 1) / nScale;
	    end
	    
	    if (childAnchor.inactive 
		or childAnchor.length ~= nLength
		   or childAnchor.angle ~= nAngle
		   or childAnchor.spin ~= nSpin
		   or childAnchor.scale ~= nScale
		   or childAnchor.relSpin) then
	       childAnchor.inactive = false;
	       childAnchor.length = nLength;
	       childAnchor.angle = nAngle;
	       childAnchor.spin = nSpin;
	       childAnchor.scale = nScale;
	       childAnchor.relSpin = false;
	       if (not child.dirty) then
		  child.dirty = true;
		  self.dirty = true;
	       end
	    end
	 end
      end
   end
end

-- Scaffold:_UpdateFollowNodeAnchor(name, node)
--   Parameters:
--     name -- The name of the following node
--     node -- The node itself
--
--   Returns:
--     anyChange -- True if the node should trigger a change event
--
--   Updates the anchor for a follow node
function Scaffold:_UpdateFollowNodeAnchor(name, node)
   local nodes = self.nodes;
   local pLoc = node.pLoc;
   local nLoc = node.nLoc;
   local loc  = node.loc;

   -- Get (or re-calculate) active child count
   local anchors = self.anchors;
   local anchor = self.anchors[name];

   -- If one or both of the points are missing, make sure this
   -- is inactive.
   if (not (nLoc and pLoc)) then
      local retChange;
      if (loc) then
	 node.loc = nil;
	 retChange = true;
      end

      local anyChange = retChange;
      if (not anchor.inactive) then
	 anchor.inactive = true;
	 anyChange = true;
      end

      if (anyChange) then
	 MarkChildrenDirty(self, nodes, node);
      end
      return retChange;
   end

   local retChange;

   -- Copy location.
   retChange = (CompareCopyLoc(nLoc,  loc,  node, "loc", "_loc")
		or retChange);

   local anyChange = retChange;

   local pX, pY, pScale, pSpin = unpack(pLoc);
   local nX, nY, nScale, nSpin = unpack(nLoc);
   local dx,dy = nX-pX, nY-pY;
   
   local scale = nScale / pScale;
   local angle = mod(deg(atan2(dx,dy)) - pSpin, 360);
   if (angle < 0) then angle = angle + 360; end
   local length = sqrt(dx*dx+dy*dy) / pScale;
   local spin;
   local relSpin = false;

   local nAnchor = anchors[node.followNode];
   if (nAnchor and nAnchor.relSpin) then
      spin =  (nSpin - pSpin) - angle;
      relSpin = true;
   else
      spin = nSpin;
   end
      
   spin = mod(spin, 360);
   if (spin < 0) then spin = spin + 360; end
	    
   if (length > 1) then
      length = 1 + (length - 1) / scale;
   end
   
   if (anchor.inactive 
       or anchor.length ~= length
	  or anchor.angle ~= angle
	  or anchor.spin ~= spin
	  or anchor.scale ~= scale
	  or anchor.relSpin ~= relSpin) then
      anchor.inactive = false;
      anchor.length = length;
      anchor.angle = angle;
      anchor.spin = spin;
      anchor.scale = scale;
      anchor.relSpin = relSpin;
      anyChange = true;
   end
   
   if (anyChange) then
      MarkChildrenDirty(self, nodes, node);
   end

   return retChange;
end

-- Scaffold:_Dispatch(name, node)
--   Parameters:
--    name - The name of the node to dispatch for
--    node - The node structure
--
--   Process event dispatch for the specified node
function Scaffold:_Dispatch(name, node)
   local disp = node.attached;
   node.dispatching = disp;
   node.changed = nil;
   if (disp) then
      local pcall,type = pcall,type; -- We're potentially calling this a LOT
      local loc = node.loc;
      local x,y,scale,spin,vState;
      if (loc) then
	 x,y,scale,spin = unpack(loc);
      end
      vState = node.visState;

      for object, func in pairs(disp) do
	 if (type(func) ~= "function") then
	    func = object[func];
	    if (type(func) ~= "function") then
	       func = nil;
	    end
	 end
	 if (func) then
	    -- Dispatch function using pcall (so errors dont
	    -- break dispatch
	    local status, errMessage;
	    status, errMessage =
	       pcall(func, object, self, name, x, y, scale, spin, vState);
	    if (not status) then
	       self:Error("Event dispatch failure on '" 
			  .. name .. "':\n" 
			     .. (errMessage or "(no error"));
	    end
	 end
      end
      node.dispatching = nil;
   end
end


-- anyChanges = Scaffold:Update()
--   Returns:
--     anyChanges -- Flag - True if any changes were made
-- 
--   Processes any dirty nodes and emits changes if necessary. If this is
--   called while Update is already running (i.e. from a change handler) then
--   it will invoke the scaffold's deferral handler and return false.
function Scaffold:Update()
   if (not self.dirty) then
      return false;
   end
   if (self.updating) then
      if (self.DeferralHandler) then
	 self:DeferralHandler();
      end
      return false;
   end

   self.updating = true;

   local nodes   = self.nodes;
   local anchors = self.anchors;
   local changes;
   local unpack = unpack;

   for _,name in ipairs(self.scan) do
      local node = nodes[name];
      if (node.dirty) then
	 node.dirty = false;
	 local nType = node.type;

	 if (node.broken) then
	    -- If this node somehow has a location, then clear it
	    if (node.loc) then
	       node.loc = nil;
	       node.changed = true;
	       MarkChildrenDirty(self, nodes, node);
	    end
	 elseif (nType == "root") then
	    -- Propagate dirtiness to children
	    MarkChildrenDirty(self, nodes, node);

	 elseif (nType == "group") then
	    local spLoc = nodes[node.startNode].loc;
	    local epLoc = nodes[node.endNode].loc;
	    local pLoc  = nodes[node.parent].loc;
	    local loc = node.loc;
	    local eLoc = node.eLoc;
	    local sLoc = node.sLoc;

	    local locChanged;

	    -- Check 'parent'
	    locChanged = (CompareCopyLoc(pLoc,  loc,  node, "loc", "_loc"));
	    
	    -- Propagate dirtiness to children
	    if (locChanged) then
	       MarkChildrenDirty(self, nodes, node);
	    end
	    
	    -- Check endpoints
	    locChanged = (CompareCopyLoc(spLoc, sLoc, node, "sLoc", "_sLoc")
			  or locChanged);
	    locChanged = (CompareCopyLoc(epLoc, eLoc, node, "eLoc", "_eLoc")
			  or locChanged);

	    if (locChanged or node.reAnchor) then
	       self:_UpdateGroupAnchors(name, node);
	       locChanged = true;
	    end

	    if (locChanged) then
	       node.changed = true;
	       changes = true;
	    end
	 elseif (nType == "node" and node.followType == "node") then
	    local npLoc = nodes[node.parent].loc;
	    local nnLoc = nodes[node.followNode].loc;

	    local pLoc = node.pLoc;
	    local nLoc = node.nLoc;

	    local nChanged;
	    local dirtyChange;

	    -- Check 'parent' and 'follow' nodes
	    nChanged = (CompareCopyLoc(npLoc,  pLoc,  
				       node, "pLoc", "_pLoc"));
	    -- If parent has moved, must dirty children
	    nChanged = (CompareCopyLoc(nnLoc,  nLoc,  
				       node, "nLoc", "_nLoc")
			   or nChanged);

	    if (nChanged) then
	       if (self:_UpdateFollowNodeAnchor(name, node)) then
		  node.changed = true;
		  changes = true;
	       end
	    end
	 else
	    -- Do anything that's not a group
	    local parent = nodes[node.parent];
	    local pLoc = parent.loc;

	    local x,y,scale,spin;
	    if (pLoc) then
	       local anchor = anchors[name];
	       
	       if (nType == "node") then
		  local followType = node.followType;
		  if (followType == "anchor") then
		     local followAnchor = anchors[node.followNode];

		     local inactive = followAnchor.inactive;
		     anchor.inactive = inactive;
		     if (not inactive) then
			anchor.angle  = followAnchor.angle;
			anchor.length = followAnchor.length;
			anchor.spin   = followAnchor.spin;
			anchor.scale  = followAnchor.scale;
			anchor.relSpin = followAnchor.relSpin;
		     end
		  end
	       end

	       x,y,scale,spin = ApplyAnchor(anchor, unpack(pLoc));
	    end

	    local loc = node.loc;
	    local isChanged;
	    if (x) then
	       if (not loc) then
		  loc = node._loc;
		  loc[1],loc[2],loc[3],loc[4] = x, y, scale, spin;
		  node.loc = loc;
		  isChanged = true;
	       elseif (loc[1] ~= x or loc[2] ~= y 
		       or loc[3] ~= scale or loc[4] ~= spin) then
		  loc[1],loc[2],loc[3],loc[4] = x, y, scale, spin;
		  isChanged = true;
	       end
	    elseif (loc) then
	       node.loc = nil;
	       isChanged = true;
	    end

	    if (isChanged) then
	       node.changed = true;
	       changes = true;
	       MarkChildrenDirty(self, nodes, node);
	    end
	 end
      elseif (node.changed) then
	 changes = true;
      end
   end

   if (changes) then
      self.changes = true;
   end

   self.dirty = nil;

   if (self.inertChanges) then
      for name,node in pairs(self.inertNodes) do
	 if (node.changes or node.dispatching) then
	    self:_Dispatch(name, node);
	 end
      end      
      self.inertChanges = nil;
   end

   if (self.changes) then
      local dirtyChanges = nil;
      for _,name in ipairs(self.scan) do
	 local node = nodes[name];
	 if (node.changed or node.dispatching) then
	    if (node.dirty) then
	       dirtyChanges = 1;
	    else
	       self:_Dispatch(name, node);
	    end
	 end
      end
      self.changes = dirtyChanges;
   end

   self.updating = nil;
   return true;
end

-- anyChanges = Scaffold:CheckUpdate()
--   Returns
--     anyChanges -- Flag - True if any changes were made
--
--   This is intended for deferred updates, and can also be used to resolve
--   problems if something catastrophic happened during a previous Update.
--   IMPORTANT: You MUST be sure that there's not another update running
--   when this is called, since it will clear the busy flag.
--   Ideally you'd call this from something like an OnUpdate handler.
function Scaffold:CheckUpdate()
   if (self.updating or self.changes) then
      self.updating = nil;
      self.dirty = 1;
   end
   if (self.dirty) then
      self:Update();
   end
end

-- needsUpdate = Scaffold:NeedsUpdate(falseIfUpdating)
--   Parameters:
--     falseIfUpdating -- Optional flag, set to true to return false when
--                        updating regardless of status.
--   Returns
--     needsUpdate -- Flag - True if there are pending update tasks
--
--   This provides a way to check if the scaffold thinks it needs to be
--   updated, though calling Update() is probably a better idea anyway since
--   it exits quickly if there's none to do.
function Scaffold:NeedsUpdate(falseIfUpdating)
   if (self.dirty) then
      if (falseIfUpdating and self.updating) then
	 return;
      end
      return true;
   end
end

-- Scaffold:Attach(nodeName, object, func or "methodName")
--    Parameters:
--     nodeName -- The name of the node to attach to
--     object   -- The object registering the handler
--     func     -- the function to call on event firing
--     methodName -- The method to call on event firing
--                   (Default "ScaffoldUpdate")
-- 
--   Registers a handler to call whenever the specified node is changed,
--   depending on whether a function or methodName is given the callback
--   is made differently.
-- 
--   if func:
-- 
--   func(object, scaffold, nodeName, locX, locY, locScale, locSpin);
--   
--   if methodName
-- 
--   object:methodName(scaffold, nodeName, locX, locY, locScale, locSpin);
-- 
--   Note that non-existant nodes can have attachments (They wont fire
--   events), and events only fire during Update() processing.
function Scaffold:Attach(name, object, func) 
   if (not func) then
      func = "ScaffoldUpdate";
   end

   local node = self.nodes[name] or self.inertNodes[name];
   if (not node) then
      node = {
	 type = "inert";
      }
      self.inertNodes[name] = node;
   end

   local attachments = node.attached;
   if (not attachments) then
      attachments = {};
      node.attached = attachments;
   elseif (node.dispatching) then
      -- We're in the process of dispatching, create a new set
      if (node.dispatching == attachments) then
	 attachments = {};
	 for k,v in pairs(node.dispatching) do
	    attachments[k] = v;
	 end
	 node.attached = attachments;
      end
   end

   attachments[object] = func;
end

-- Scaffold:Detach(nodeName, object);
--    Parameters:
--     nodeName -- The name of the node to attach to
--     object   -- The object unregistering the handler
-- 
--    Unregisters a previously attached handler.
function Scaffold:Detach(name, object) 
   local node = self.nodes[name] or self.inertNodes[name];
   if (not node) then
      return;
   end

   local attachments = node.attached;
   if (not attachments) then
      return;
   end
   
   if (node.dispatching) then
      -- We're in the process of dispatching, create a new set
      if (node.dispatching == attachments) then
	 attachments = {};
	 for k,v in pairs(node.dispatching) do
	    attachments[k] = v;
	 end
	 node.attached = attachments;
      end
   end

   if (attachments[object]) then
      attachments[object] = nil;

      if ((not self.updating) and (node.type == "inert")) then
	 if (next(attachments) == nil) then
	    node.attached = nil;      -- We can discard this object
	    if (not node.dispatching) then
	       self.inertNodes[name] = nil;
	    end
	 end
      end
   end
end

-- exists = Scaffold:NodeExists(nodeName)
--   Parameters:
--    nodeName -- the name of a node
-- 
--   Returns:
--    exists -- Flag -- True if the node exists
function Scaffold:NodeExists(nodeName)
   return (self.nodes[nodeName] and true) or nil;
end

-- num = Scaffold:GetNumNodes()
--   Returns:
--    num -- How many nodes are in the scaffold
function Scaffold:GetNumNodes()
   return table.getn(self.scan);
end

-- num = Scaffold:GetNodeName(idx)
--   Parameters:
--    idx -- Which node to get the name for
--
--   Returns:
--    name -- The name of the selected node
function Scaffold:GetNodeName(idx)
   return self.scan[idx];
end

-- nodeType = Scaffold:GetType(nodeName);
--   Parameters:
--     nodeName -- The name of a node
-- 
--   Returns:
--     nodeType -- The type of the node nil, "node", "member", "group"
--     isBroken -- True if the node is broken, nil otherwise
function Scaffold:GetType(nodeName)
   local node = self.nodes[nodeName];
   return (node and node.type),(node and node.broken and true) or nil;
end

-- x,y,scale,spin = Scaffold:GetLocation(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     x,y -- Location of the node (nil if inactive/broken/unshowable)
--     scale -- Scale of the node (nil if inactive/broken/unshowable)
--     spin -- Spin (nil if inactive/broken/unshowable)
function Scaffold:GetLocation(nodeName)
   local node = self.nodes[nodeName];
   if (node) then
      local loc = node.loc;
      if (loc) then
	 return unpack(loc)
      end
   end
end

-- parentName = Scaffold:GetParent(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     parentName - The node's parent, or nil if it has no parent
function Scaffold:GetParent(nodeName)
   local node = self.nodes[nodeName];
   return node and node.parent;
end

-- anchor, isDefault = Scaffold:GetAnchor(nodeName[, returnTable])
--   Parameters:
--     nodeName -- The name of a node
--     returnTable -- Table in which to return the results (only the normal
--                    anchor fields will be reset), a new table will be
--                    created if this is nil
--   Returns:
--     anchor -- The anchor result, or nil if the node does not exist or has
--               no anchor
--     isDefault -- True if the anchor is the node's default
function Scaffold:GetAnchor(name, ret)
   local anchor = self.anchors[name];
   if ((not anchor) or anchor.inactive) then
      return nil;
   end
   if (not ret) then
      ret = {};
   end
   
   CopyAnchor(anchor, ret);
   ret.derived = anchor.derived;
   return ret;
end

-- canSetAnchor = Scaffold:CanSetAnchor(nodeName)
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--    canSetAnchor -- Flag -- True if the anchor can be updated, nil otherwise.
function Scaffold:CanSetAnchor(nodeName)
   local anchor = self.anchors[nodeName];
   if (anchor and (not anchor.derived)) then
      return true;
   end
end

-- canSetAnchor = Scaffold:SetAnchor(nodeName, newAnchor)
--   Parameters:
--     nodeName -- The name of a node
--     newAnchor -- The desired new anchor (nil means 'use node's default)
--   Returns:
--    canSetAnchor -- Flag -- True if the anchor could be (and thus was)
--                            updated, nil otherwise.
function Scaffold:SetAnchor(nodeName, anchorSpec)
   local anchor = self.anchors[nodeName];
   if ((not anchor) or anchor.derived) then
      return;
   end

   local isOverride = true;

   if (anchorSpec == nil) then
      anchorSpec = self.nodes[node].defaultAnchor or EMPTY_TABLE;
      isOverride = false;
   end

   local scale  = tonumber(anchorSpec.scale) or 1.0;
   local length = tonumber(anchorSpec.length) or 0.0;
   local angle  = tonumber(anchorSpec.angle) or 0;
   local spin   = tonumber(anchorSpec.spin) or 0;
   local relSpin;
   if (anchorSpec.relSpin) then relSpin = true; end

   anchor.isOverride = isOverride;

   if (scale  == anchor.scale
       and length  == anchor.length
	  and angle  == anchor.angle
	  and spin    == anchor.spin
	  and relSpin == anchor.relSpin) then
      return true;
   end

   local node = self.nodes[nodeName];

   anchor.scale = scale;
   anchor.length = length;
   anchor.angle = angle;
   anchor.spin = spin;
   anchor.relSpin = relSpin;

   if (not node.dirty) then
      node.dirty = true;
      self.dirty = true;
   end

   return true;
end

-- state = Scaffold:GetState(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     state -- String -- Current visibility state of the node, nil if no node.
function Scaffold:GetState(name)
   local node = self.nodes[name];
   if (not node) then
      return nil;
   end
   return node.visState or (node.type == "member" and "HIDDEN") or "SHOWN";
end

-- state = Scaffold:GetDefaultState(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     state -- String -- Current default visibility state of the node,
--                        nil if no node.
function Scaffold:GetDefaultState(name)
   return self.defaultVisibilityManager:GetState(name);
end

-- oldState = ScaffoldVisibility:SetDefaultState(nodeName, newState);
--   Parameters:
--     nodeName -- The name of a node.
--     newState -- Desired new state (default is node default)
-- 
--   Returns:
--    oldState -- String -- The old state, nil if the node did not exist.
function Scaffold:SetDefaultState(name, newState)
   return self.defaultVisibilityManager:SetState(name, newState);
end

-- canSetActive = Scaffold:_SetState(nodeName, newState);
--   Parameters:
--     nodeName -- The name of a node.
--     newState -- Desired visibilitty state
-- 
--   Returns:
--    canSetActive -- Flag -- True if the node could be (and thus was) set
--                            active/inactive, nil otherwise.
function Scaffold:_SetState(name, newState)
   local nodes = self.nodes;
   local node = nodes[name];
   if (not node) then
      return nil;
   end

   local oldState = node.visState;

   if (not newState) then
      newState = (node.type == "member" and "HIDDEN") or "SHOWN";
   end

   if (oldState == newState) then
      return true;
   end

   node.visState = newState;
   node.changed = 1;
   self.dirty = 1;
   self.changes = 1;

   if (node.type ~= "member") then
      return true;
   end

   if ((oldState ~= "HIDDEN") and (newState ~= "HIDDEN")) then
      return true;
   end

   -- Member node has transitioned visibility
   local group = nodes[node.parent];
   group.activeCount = nil; -- Reset active count.
   group.reAnchor = true; -- Force anchor re-calculation
   if (not group.dirty) then
      group.dirty = true;
   end
   if (not node.dirty) then
      node.dirty = true;
   end
   return true;
end

-- number = Scaffold:GetNumGroupMembers(nodeName)
--   Parameters:
--     nodeName -- The name of a node
--
--   Returns
--     number -- The number of members the node has, if a group, else nil.
function Scaffold:GetNumGroupMembers(name)
   local node = self.nodes[name];
   if (not (node and node.type == "group")) then
      return;
   end
   return table.getn(node.members);
end

-- memberName = Scaffold:GetGroupMember(nodeName, idx)
--   Parameters:
--     nodeName -- The name of a node
--     idx -- The member index
--
--   Returns
--     memberName -- The specified member's name, or nil if args invalid
function Scaffold:GetGroupMember(name, idx)
   local node = self.nodes[name];
   if (not (node and node.type == "group")) then
      return;
   end
   return node.members[idx];
end

-- groupSpec = Scaffold:GetGroupSpec(nodeName[, returnTable])
--   Parameters:
--     nodeName -- The name of a node
--     returnTable -- Table in which to return the results (only the normal
--                    groupSpec fields will be reset), a new table will be
--                    created if this is nil
--   Returns:
--     groupSpec -- The groupSpec result, or nil if the node does not exist or
--                  has no groupSpec.
function Scaffold:GetGroupSpec(name, ret)
   local group = self.nodes[name];
   if ((not group) or (group.type ~= "group")) then
      return nil;
   end
   if (not ret) then
      ret = {};
   end

   CopyGroupSpec(group, ret);
   return ret;
end

-- canSetGroupSpec = Scaffold:CanSetGroupSpec(nodeName)
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--    canSetGroupSpec -- Flag -- True if the groupSpec can be updated, nil
--                               otherwise.
function Scaffold:CanSetGroupSpec(nodeName)
   local node = self.node[nodeName];
   if (node and (node.type == "group")) then
      return true;
   end
end

-- canSetGroupSpec = Scaffold:SetGroupSpec(nodeName, newGroupSpec)
--   Parameters:
--     nodeName -- The name of a node
--     newGroupSpec -- The desired new groupSpec.
--   Returns:
--    canSetGroupSpec -- Flag -- True if the groupSpec could be (and thus was)
--                               updated, nil otherwise.
function Scaffold:SetGroupSpec(name, spec)
   local group = self.nodes[name];
   if ((not group) or (group.type ~= "group")) then
      return;
   end
   local size = tonumber(spec.size) or 0;
   local spread = spec.spread or "none";
   local weight = tonumber(spec.weight) or 0;
   local spinReverse = (spec.spinReverse and true) or nil;

   if (group.size == size
       and group.spread == spread
	  and group.weight == weight
	  and group.spinReverse == spinReverse) then
      return true;
   end

   group.size        = size;
   group.spread      = spread;
   group.weight      = weight;
   group.spinReverse = spinReverse;
   if (size == 0) then
      group.effSize = table.getn(group.members);
   else
      group.effSize = size;
   end

   group.reAnchor    = true;
   if (not group.dirty) then
      group.dirty = true;
      self.dirty = true;
   end
   return true;
end

-- followNodeName = Scaffold:GetFollowNode(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     followNodeName -- If this node has a follow node, it is returned
--                       otherwise nil
function Scaffold:GetFollowNode(nodeName)
   local node = self.nodes[nodeName];
   if (node and node.type=="node") then
      return node.followNode;
   end
end

-- node1,node2 = Scaffold:GetDependencies(nodeName);
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--     The names of the node's dependencies (max 2)
function Scaffold:GetDependencies(nodeName)
   local node = self.nodes[nodeName];
   if (not node) then
      return;
   end
   local nType = node.type;
   if (nType == "root") then
      return; 
   elseif (nType == "member") then
      return node.parent;
   elseif (nType == "node") then
      if (node.followNode) then
	 return node.parent, node.followNode;
      end
      return node.parent;
   elseif (nType == "group") then
      return node.startNode, node.endNode;
   end
end

-- canSetRootLocation = Scaffold:CanSetRootLocation(nodeName)
--   Parameters:
--     nodeName -- The name of a node.
-- 
--   Returns:
--    canSetRootLocation -- Flag -- True if the location can be updated,
--                                   nil otherwise.
function Scaffold:CanSetRootLocation(nodeName)
   local node = self.nodes[nodeName];
   if (node and (node.type == "root")) then
      return true;
   end
end

-- canSetRootLocation = Scaffold:SetRootLocation(nodeName, x, y, scale, spin)
--   Parameters:
--     nodeName -- The name of a node
--     x,y -- The root's location (nil for invalid root)
--     scale -- The root's scale (nil for invalid root)
--     spin -- The root's spin (nil for invalid root)
--   Returns:
--    canSetRootLocation -- Flag -- True if the location could be (and thus
--                                  was) updated, nil otherwise.
function Scaffold:SetRootLocation(nodeName, x, y, scale, spin)
   local node = self.nodes[nodeName];
   if (not (node and (node.type == "root"))) then
      return;
   end

   node.override = true;
   local loc = node.loc;
   if (x) then
      x = tonumber(x) or 0;
      y = tonumber(y) or 0;
      scale = tonumber(scale) or 1;
      spin = tonumber(spin) or 0;

      if (not loc) then
	 loc = node._loc;
	 node.loc = loc;
      else
	 if ((loc[1] == x) and (loc[2] == y)
	     and (loc[3] == scale) and (loc[4] == spin)) then
	    -- No change
	    return true;
	 end
      end
      loc[1] = x;
      loc[2] = y;
      loc[3] = scale;
      loc[4] = spin;
      if (not node.dirty) then
	 node.dirty = true;
	 self.dirty = true;
      end
      return true;
   end

   if (not loc) then
      return true;
   end
   
   node.loc = nil;
   if (not node.dirty) then
      node.dirty = true;
      self.dirty = true;
   end
   return true;
end

function Scaffold:GetVisibilityManager()
   return self.visibilityManager;
end

function Scaffold:CreateVisibilityManager(set)
   local mgr = StatRingsScaffoldVisibilityManager_New(self);
   if (set) then
      self:SetVisibilityManager(mgr);
   end
   return mgr;
end

function Scaffold:SetVisibilityManager(visManager)
   if (not visManager) then
      visManager = self.defaultVisibilityManager;
   end
   if (visManager == self.visibilityManager) then
      return;
   end
   if (visManager.scaffold ~= self) then
      self:Error("Attempt to use incompatible visibility manager!");
      return;
   end

   self.visibilityManager = visManager;
   for _, name in ipairs(self.scan) do
      self:_SetState(name, visManager.states[name]);
   end
end

-- rootFrame = Scaffold:GetRootFrame()
--   Returns:
--     rootFrame - The root frame for this scaffold
function Scaffold:GetRootFrame()
   return self.rootFrame;
end

function StatRingsScaffold_New(rootFrame)
   local ret = {
      rootFrame = rootFrame;
      nodes = { };
      inertNodes = { };
      anchors = { };
      scan = { };
   };

   setmetatable(ret, { __index = Scaffold });

   local visMgr = StatRingsScaffoldVisibilityManager_New(ret);
   ret.defaultVisibilityManager = visMgr;
   ret.visibilityManager = visMgr;

   return ret;
end

local ScaffoldVisibility = {};

-- isScaffold = ScaffoldVisibility:IsForScaffold(scaffold);
--   Parameters:
--     scaffold -- The scaffold to test against.
-- 
--   Returns:
--     isScaffold -- Flag -- True if the given scaffold is this
--                           manager's scaffold.
function ScaffoldVisibility:IsForScaffold(scaffold)
   return (scaffold == self.scaffold) or nil;
end

-- state = ScaffoldVisibility:GetState(nodeName, unknownOk);
--   Parameters:
--     nodeName -- The name of a node.
--     unknownOk -- Set to nil if unknown nodes with a state get that
--                  state returned;
-- 
--   Returns:
--     state -- String -- The state of the node (nil for unknown nodes
--                        with no state, or (by default) all unknown nodes)
function ScaffoldVisibility:GetState(name, unknownOk)
   local node = self.scaffold.nodes[name];
   if (not node) then
      if (unknownOk) then
	 return self.states[name];
      else
	 return nil; 
      end
   end
   local state = self.states[name];
   return state or (node.type == "member" and "HIDDEN") or "SHOWN";
end

-- oldState = ScaffoldVisibility:SetState(nodeName, newState);
--   Parameters:
--     nodeName -- The name of a node.
--     newState -- Desired new state (default is node default)
-- 
--   Returns:
--    oldState -- String -- The old state, nil if the node did not exist.
--                          (Nil is returned regardless of previous node state)
function ScaffoldVisibility:SetState(name, newState)
   local scaffold = self.scaffold;
   local nodes = scaffold.nodes;
   local node = nodes[name];
   local states = self.states;
   if (not node) then
      states[name] = newState;
      return;
   end
   local nodeType = node.type;
   local nState = newState or (nodeType == "member" and "HIDDEN") or "SHOWN";
   local oState = states[name] or (nodeType == "member" and "HIDDEN") 
      or "SHOWN";
   if (nState == oState) then
      return oState;
   end

   states[name] = nState;
   if (scaffold.visibilityManager == self) then
      scaffold:_SetState(name, nState);
   end

   return oState;
end

function StatRingsScaffoldVisibilityManager_New(scaffold)
   local ret = {
      scaffold = scaffold;
      states = {};
   };

   setmetatable(ret, { __index = ScaffoldVisibility });

   return ret;
end

-----------------------------------------------------------------------------
StatRings_NewScaffoldSpec = {
   nodes = {
      player = {
	 anchor = { };
      };

      pet = {
	 parent = "player";
	 anchor = { scale=0.4; angle=180; length=0.6; 
	            relSpin=true; spin=180; };
      };

      target = {
	 parent = "player";
	 anchor = { scale=0.9; relSpin=true;};
      };

      partyStart = {
	 parent = "player";
	 anchor = { scale=0.25; angle=-37.5; length=2.1; 
	            relSpin=true; spin=37.5};
      };

      partyEnd = {
	 parent = "player";
	 anchor = { scale=0.25; angle=37.5; length=2.1;
	            relSpin = true;  spin=-37.5};
      };
      
      party = {
	 size = 4; -- Technically optional based on members array
	 spread = "none"; -- Default 'none', can be 'open','fill'
	 weight = 0.5; -- Anywhere from 0 to 1

	 members = {
	    "party1",
	    "party2",
	    "party3",
	    "party4",
	 }
      };

      partypet = {
	 parent = "player";
	 followType = "node";
	 followNode = "pet"
      };

      partypet1 = {
	 parent = "party1";
	 followType = "anchor";
	 followNode = "partypet";
      };

      partypet2 = {
	 parent = "party2";
	 followType = "anchor";
	 followNode = "partypet";
      };

      partypet3 = {
	 parent = "party3";
	 followType = "anchor";
	 followNode = "partypet";
      };

      partypet4 = {
	 parent = "party4";
	 followType = "anchor";
	 followNode = "partypet";
      };

      comboRoot = {
	 parent = "player";
	 anchor = { scale=1; length=0; };
      };

      comboStart = {
	 parent = "comboRoot";
	 anchor = { scale=0.09; length=0.94; angle=204};
      };

      comboEnd = {
	 parent = "comboRoot";
	 anchor = { scale=0.09; length=0.94; angle=156};
      };

      combo = {
	 --size = 6;

	 spread = "none"; -- Default 'none', can be 'open','fill'
	 weight = 0.5; -- Anywhere from 0 to 1
	 --spread = "fill";
	 --weight = 0;
	 --spinReverse = true;

	 members = {
	    "combo1",
	    "combo2",
	    "combo3",
	    "combo4",
	    "combo5",
	 };
      };
   };
};

